/*
  select.c - Brute force approach to solving Microcosm.

  Jason Hood, 17 to 25 January, 2009.

  3 March, 2009:
    speed optimisation - use an array instead of modulus.

  4 March, 2009:
    read lines from file.

  9 to 15 August, 2014:
    use program name or first argument as base for lines/progress/found names:
      e.g. select1.exe -> select1-lines, select1-progress, select1-found;
	   select test -> test-lines, test-progress, test-found;
    make the progress file read-only on completion;
    save progress at most once every minute;
    use SSE2 to process 16 counters at a time;
    -h option to run hidden;
    index by four characters, increase to 12-letter words.

  18 October, 2014:
    show the window upon completion;
    ignore duplicated numbers (allowing something like "2 4 6 1-16");
    allow key number to have a suffix:
      'k' for a key inside a key;
      'p' for a "plain" key;
      'r' for a keyring key.

  25 October, 2014:
    show the window upon finding a message;
    hide the window with Ctrl+Del, removing existing messages;
    estimate progress based on the progress of the session.

  30 October, 2014:
    show progress as a percentage.

  26 April, 2022:
    make the progress file empty on completion, not read-only;
    restart if the lines file is newer than the progress file;
    display the message key as page & suffix.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <io.h>
#include <conio.h>


#define COUNTERS 20
#define PAGES	 13
#define LINES	 16
#define KEYS	 21		// 16 keys on 13 pages, plus 5 on the keyring
#define CODES	 14		// a line from each page, plus the key


typedef char v16qi __attribute__((vector_size(16)));

struct dwvector_t
{
  DWORD64 n;
  v16qi   v;
};
typedef struct dwvector_t DWVECTOR;

union counter_t
{
  DWORD64 m;			// the first four counters
  WORD	  w[4];
  struct
  {
    DWORD dummyn[3];		// to make n directly precede v
    DWORD n;			// the first four counters
    v16qi v;			// the remaining 16
  };
  struct
  {
    char dummyc[12];
    unsigned char c[COUNTERS+1]; // an extra for a sentinel (terminating space)
  };
};
typedef union counter_t COUNTER;

int   n, N[CODES];
COUNTER m;
DWORD64 M[CODES];
int   p[CODES], L[CODES][KEYS+1]; // one more for terminator
char  map[10*26];		// map counter to character

const int mod20[] =		// create a mod 20 map
{
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
  0, 1, 2, 3, 4, 5,		// 19 (last counter) + 26 (longest line) = 45
};

const int mod26[] =		// create a mod 26 map
{
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,
};


int   hidden;
DWORD progress;
char  lines_name[MAX_PATH], found_name[MAX_PATH], progress_name[MAX_PATH];

#include "select.h"
#include "keys.h"


// Create the word list.
// Assumptions: the word list is sorted;
//		it contains only lowercase letters 'a' to 'y';
//		words are between two and twelve letters;
//		LF line ending.

union charnum_t
{
  long long n;
  unsigned char c[8];
};
typedef union charnum_t CHARNUM;

struct words_t
{
  int last;		// index of the last word
  CHARNUM words[0];
};
typedef struct words_t WORDS;

WORDS* start[26][32][32][32];


void print_message( WORDS* );


void build( void )
{
  union
  {
    DWORD n;
    char c[16];
  } word, next;
  int  i, j;

  // Open binary to ensure no problems with seeking.
  FILE* file = fopen( "wordlist.12", "rb" );
  if (file == NULL)
  {
    puts( "Could not open \"wordlist.12\"!" );
    exit( 1 );
  }

  for (;;)
  {
    long pos = ftell( file );
    if (fgets( word.c, sizeof(word), file ) == NULL)
      break;
    if (word.c[2] == '\n')
    {
      start[0][0][word.c[1]-'a'+1][word.c[0]-'a'+1] = (WORDS*)1;
    }
    else if (word.c[3] == '\n')
    {
      start[0][word.c[2]-'a'+1][word.c[1]-'a'+1][word.c[0]-'a'+1] = (WORDS*)1;
    }
    else
    {
      // Find the number of words with the same first four letters.
      int num = 1;
      while (fgets( next.c, sizeof(next), file ) && next.n == word.n)
	++num;
      fseek( file, pos, SEEK_SET );
      WORDS* w = malloc( sizeof(WORDS) + num * sizeof(CHARNUM) );
      start[word.c[3]-'a'+1][word.c[2]-'a'+1][word.c[1]-'a'+1][word.c[0]-'a'+1] = w;
      w->last = num - 1;
      for (i = 0; i < num; ++i)
      {
	fgets( word.c, sizeof(word), file );
	w->words[i].n = 0;
	for (j = 4; word.c[j] != '\n'; ++j)
	  w->words[i].c[7-(j-4)] = word.c[j] - 'a' + 1;
      }
    }
  }

  fclose( file );
}


void save_progress( void )
{
  FILE* file;
  int	j;

  file = fopen( progress_name, "w" );
  for (j = 0; j < CODES; ++j)
    fprintf( file, "%2d ", p[j] );
  fprintf( file, "%lu\n", progress );
  fclose( file );
}


// Control-break handler - save progress & exit.
BOOL WINAPI Handler( DWORD type )
{
  save_progress();
  ExitProcess( 0 );
  return TRUE;
}


int main( int argc, char* argv[] )
{
  int	i, j, k;
  char	buf[256];
  char* ptr;
  DWORD t, tt, ts;
  DWORD st, sp;
  DWORD count = 1;
  DWORD pr;
  BYTE	used[KEYS+5];
  int	n1, n2;
  WIN32_FILE_ATTRIBUTE_DATA lt, pt;
  FILE* file;

  if (argc > 1 && (strcmp( argv[1], "--help" ) == 0 ||
		   strcmp( argv[1], "/?" ) == 0))
  {
    puts( "Microcosm by Hal Gashtan.\n"
	  "This program by Jason Hood <jadoxa@yahoo.com.au>.\n"
	  "http://ucosm.adoxa.vze.com/\n"
	  "\n"
	  //"Wordlist Copyright (c) J Ross Beresford 1993-2000. All Rights Reserved.\n"
	  "Wordlist from 12dicts, by Alan Beale.\n"
	  "Modified to be more suited to Microcosm." );
    return 0;
  }

  build();

  for (j = 0; j < 10*26; j += 26)
    memcpy( map+j, " abcdefghijklmnopqrstuvwxy", 26 );

  i = 1;
  if (argc > 1 && strcmp( argv[1], "-h" ) == 0)
  {
    i = 2;
    hidden = 1;
  }
  ptr = (argv[i]) ? argv[i] : argv[0];
  i = strlen( ptr );
  if (i > 4 && _stricmp( ptr + i - 4, ".exe" ) == 0)
  {
    i -= 4;
    ptr[i] = '\0';
  }
  sprintf( lines_name, "%s-lines", ptr );
  sprintf( progress_name, "%s-progress", ptr );
  sprintf( found_name, "%s-found", ptr );

  if (GetFileAttributesEx( lines_name, GetFileExInfoStandard, &lt ) &&
      GetFileAttributesEx( progress_name, GetFileExInfoStandard, &pt ) &&
      CompareFileTime( &lt.ftLastWriteTime, &pt.ftLastWriteTime ) > 0)
    remove( progress_name );

  file = fopen( progress_name, "r" );
  if (file)
  {
    int c = getc( file );
    if (c == EOF)
    {
      puts( "Completed!" );
      if (GetConsoleProcessList( &pr, 1 ) == 1)
	_getch();
      return 0;
    }
    ungetc( c, file );
    for (j = 0; j < CODES; ++j)
      fscanf( file, "%d", &p[j] );
    fscanf( file, "%lu", &progress );
    fclose( file );
  }

  file = fopen( lines_name, "r" );
  if (!file)
  {
    printf( "Place your line selections in \"%s\".\n", lines_name );
    return 1;
  }
  for (j = 0; j < PAGES; ++j)
  {
    *buf = '\0';
    fgets( buf, sizeof(buf), file );
    for (ptr = buf; isspace( *ptr ); ++ptr) ;
    if (!isdigit( *ptr ))
    {
      for (i = 0; i < LINES; ++i)
	L[j][i] = i;
    }
    else
    {
      memset( used, FALSE, sizeof(used) );
      for (i = 0; *ptr; ++ptr)
      {
	while (isspace( *ptr ))
	  ++ptr;
	if (*ptr == ';' || *ptr == '#')
	  break;
	n1 = n2 = strtol( ptr, &ptr, 10 ) - 1;
	if (n1 < 0 || n1 >= LINES)
	{
	  printf( "Invalid line for page %d.\n", j + 1 );
	  return 1;
	}
	if (*ptr == '-')
	{
	  n2 = strtol( ptr + 1, &ptr, 10 ) - 1;
	  if (n2 < n1 || n2 >= LINES)
	  {
	    printf( "Invalid range for page %d.\n", j + 1 );
	    return 1;
	  }
	}
	for (; n1 <= n2; ++n1)
	{
	  if (!used[n1])
	  {
	    used[n1] = TRUE;
	    L[j][i++] = n1;
	  }
	}
      }
    }
    L[j][i] = -1;
    if (j < 6)
      count *= i;
  }

  *buf = '\0';
  fgets( buf, sizeof(buf), file );
  for (ptr = buf; isspace( *ptr ); ++ptr) ;
  if (!isdigit( *ptr ))
  {
    for (i = k = 0; k < 16; ++k)
      if (k != 10)		// Page 9 key has been found
	L[j][i++] = k;
  }
  else
  {
    memset( used, FALSE, sizeof(used) );
    for (i = 0; *ptr; ++ptr)
    {
      while (isspace( *ptr ))
	++ptr;
      if (*ptr == ';' || *ptr == '#')
	break;
      n1 = n2 = strtol( ptr, &ptr, 10 ) - 1;
      if (*ptr == 'k' || *ptr == 'p')
      {
	if (n1 < 0 || n1 >= PAGES)
	{
	  printf( "Invalid key.\n" );
	  return 1;
	}
	n1 = n2 = (*ptr == 'k') ? key_k[n1] : key_p[n1];
	++ptr;
      }
      else if (*ptr == 'r')
      {
	if (n1 < 0 || n1 >= 5)
	{
	  printf( "Invalid key.\n" );
	  return 1;
	}
	n2 = n1 += 16;
	++ptr;
      }
      else
      {
	if (n1 < 0 || n1 >= KEYS)
	{
	  printf( "Invalid key.\n" );
	  return 1;
	}
	if (*ptr == '-')
	{
	  n2 = strtol( ptr + 1, &ptr, 10 ) - 1;
	  if (n2 < n1 || n2 >= KEYS)
	  {
	    printf( "Invalid key range.\n" );
	    return 1;
	  }
	}
      }
      for (; n1 <= n2; ++n1)
      {
	if (!used[n1])
	{
	  used[n1] = TRUE;
	  L[j][i++] = n1;
	}
      }
    }
  }
  L[j][i] = -1;
  fclose( file );

  if (hidden)
    ShowWindow( GetConsoleWindow(), SW_HIDE );

  SetConsoleCtrlHandler( Handler, TRUE );
  SetPriorityClass( GetCurrentProcess(), IDLE_PRIORITY_CLASS );

  st = ts = tt = GetTickCount();
  sp = progress;

#define PAGE( c ) \
  N[c] = n;\
  M[c] = m.m;\
  for (; L[c][p[c]] >= 0; ++p[c])\
  {\
    if (c < PAGES)\
    {\
      m.m += pages[n][c][L[c][p[c]]].n;\
      n = mod20[n + lengths[c][L[c][p[c]]]];\
    }\
    else\
    {\
      m.m += keys[n][L[13][p[13]]].n;\
    }

#define NEXT( c ) \
    n = N[c];\
    m.m = M[c];\
  }\
  p[c] = 0;


  PAGE(  0 )
  PAGE(  1 )
  PAGE(  2 )
  PAGE(  3 )
  PAGE(  4 )
  PAGE(  5 )
  PAGE(  6 )
  PAGE(  7 )
  PAGE(  8 )
  PAGE(  9 )
  PAGE( 10 )
  PAGE( 11 )
  PAGE( 12 )
  PAGE( 13 )		// the key

  m.c[0] = mod26[m.w[0]];
  if (m.c[0] != 0)
  {
    m.c[1] = mod26[m.w[1]];
    if (m.c[1] == 0)
    {
      if (m.c[0] == 1 || m.c[0] == 9)
      {
	m.c[2] = mod26[m.w[2]];
	m.c[3] = mod26[m.w[3]];
	if (m.c[2] != 0 && m.c[3] != 0)
	  print_message( (WORDS*)2 );
      }
    }
    else
    {
      m.c[2] = mod26[m.w[2]];
      if (m.c[2] == 0)
      {
	m.c[3] = mod26[m.w[3]];
	if (m.c[3] != 0 && start[0][0][m.c[1]][m.c[0]])
	  print_message( (WORDS*)3 );
      }
      else
      {
	m.c[3] = mod26[m.w[3]];
	if (m.c[3] == 0)
	{
	  if (start[0][m.c[2]][m.c[1]][m.c[0]])
	    print_message( (WORDS*)4 );
	}
	else
	{
	  if (start[m.c[3]][m.c[2]][m.c[1]][m.c[0]])
	    print_message( start[m.c[3]][m.c[2]][m.c[1]][m.c[0]] );
	}
      }
    }
  }

  NEXT( 13 )
  NEXT( 12 )
  NEXT( 11 )
  NEXT( 10 )
  NEXT(  9 )
  NEXT(  8 )
  NEXT(  7 )
  NEXT(  6 )

  ++progress;

  t = GetTickCount();
  if (t - ts >= 60000)
  {
    save_progress();
    ts = t;
  }

  if (!hidden)
  {
    if (t - tt >= 600000)
    {
      // Took too long, probably went to sleep - reset.
      st = tt = t;
      sp = progress;
    }
    else if (t - tt >= 10000)
    {
      int h, m, s;
      DWORD tim;
      printf( "\rEstimated time remaining: " );
      tim = (DWORD)((long long)(count - progress) * (t - st)
			       / (progress - sp) / 1000);
      h = tim / 3600;
      m = (tim % 3600) / 60;
      s = tim % 60;
      if (h)
	printf( "%dh", h );
      if (m)
      {
	if (h)
	  putchar( ' ' );
	printf( "%dm", m );
      }
      if (s)
      {
	if (h || m)
	  putchar( ' ' );
	printf( "%ds", s );
      }
      printf( " (%.3f%%)         \b\b\b\b\b\b\b\b\b",
	      progress * 100.0 / count );

      if (_kbhit() && _getch() == 224 && _getch() == 147)
      {
	HANDLE con;
	CONSOLE_SCREEN_BUFFER_INFO csbi;
	COORD pos;
	DWORD len;

	ShowWindow( GetConsoleWindow(), SW_HIDE );
	hidden = 1;

	con = GetStdHandle( STD_OUTPUT_HANDLE );
	GetConsoleScreenBufferInfo( con, &csbi );
	pos.X = pos.Y = 0;
	FillConsoleOutputCharacter( con, ' ', (csbi.dwCursorPosition.Y + 1)
					      * csbi.dwSize.X, pos, &len );
	SetConsoleCursorPosition( con, pos );
      }

      tt = t;
    }
  }

  NEXT(  5 )
  NEXT(  4 )
  NEXT(  3 )
  NEXT(  2 )
  NEXT(  1 )
  NEXT(  0 )

  fclose( fopen( progress_name, "w" ) );
  puts( "\rCompleted!                            " );
  if (hidden)
  {
    ShowWindow( GetConsoleWindow(), SW_SHOWNOACTIVATE );
    if (GetConsoleProcessList( &pr, 1 ) == 1)
      _getch();
  }

  return 0;
}


void print_message( WORDS* w )
{
  CHARNUM cn;
  int	  mid;
  FILE*   file;
  int	  i, j, k;
  unsigned char c[4];

  k = 0;
  memset( m.c + 4, 0, 16 );
  for (i = 0; i < PAGES; ++i)
  {
    // Need to mod the counters in order to prevent byte overflow.
    if (i == 6)
      for (j = 4; j < COUNTERS; ++j)
	m.c[j] = mod26[m.c[j]];
    m.v += pages[k][i][L[i][p[i]]].v;
    k = mod20[k + lengths[i][L[i][p[i]]]];
  }
  m.v += keys[k][L[13][p[13]]].v;

  // Determine the validity of the message.
  // Assumptions: no message will start with a space;
  //		  there will not be consecutive spaces (except for trailing);
  //		  there will be no words greater than twelve letters;
  //		  all words will exist in the list.

  if ((DWORD_PTR)w > 4)
  {
    i = 3;
    goto rest;
  }
  for (i = (int)(INT_PTR)w; i < COUNTERS; ++i)
  {
    c[0] = mod26[m.c[i]];
    if (c[0] == 0)
    {
      while (++i < COUNTERS)
      {
	if (mod26[m.c[i]] != 0)
	  return;
      }
      break;
    }
    if (i == COUNTERS-1)
    {
      if (c[0] == 9)
	continue;
      return;
    }
    c[1] = mod26[m.c[++i]];
    if (c[1] == 0)
    {
      if (c[0] == 1 || c[0] == 9)
	continue;
      return;
    }
    c[2] = mod26[m.c[++i]];
    if (c[2] == 0)
    {
      if (!start[0][0][c[1]][c[0]])
	return;
      continue;
    }
    c[3] = mod26[m.c[++i]];
    if (c[3] == 0)
    {
      if (!start[0][c[2]][c[1]][c[0]])
	return;
      continue;
    }
    w = start[c[3]][c[2]][c[1]][c[0]];
    if (!w)
      return;
  rest:
    j = 8;	// maximum number of additional letters
    cn.n = 0;
    while ((c[0] = mod26[m.c[++i]]) != 0)
    {
      if (--j < 0)
	return;
      cn.c[j] = c[0];
    }
    // Search the list for the word.
    j = 0;
    k = w->last;
    do
    {
      mid = (j + k) / 2;
      if (cn.n == w->words[mid].n)
	goto next;
      if (cn.n < w->words[mid].n)
	k = mid - 1;
      else
	j = mid + 1;
    } while (j <= k);
    return;
  next: ;
  }

  if (hidden)
  {
    ShowWindow( GetConsoleWindow(), SW_SHOWNOACTIVATE );
    hidden = 0;
  }

  putchar( '\r' );
  for (i = 0; i < CODES - 1; ++i)
    printf( "%2d ", L[i][p[i]] + 1 );
  printf( "%3s : ", key_s[L[i][p[i]]] );
  for (i = 0; i < COUNTERS; ++i)
    putchar( map[m.c[i]] );
  putchar( '\n' );

  file = fopen( found_name, "a" );
  for (i = 0; i < CODES; ++i)
    fprintf( file, "%2d ", L[i][p[i]] );
  putc( ':', file );
  putc( ' ', file );
  for (i = 0; i < COUNTERS; ++i)
    putc( map[m.c[i]], file );
  putc( '\n', file );
  fclose( file );
}
